ISO_DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
SERVER_ISO_DATE_FORMAT = "yyyy-MM-ddTHH:mm:ssZ"

SIZE_MISMATCH = "SIZE_MISMATCH"
COLOR_MISMATCH = "COLOR_MISMATCH "
SAMPLE_MISMATCH = "SAMPLE_MISMATCH"
EXCESS_GLUE = "EXCESS_GLUE"
WEEK_DEGUMMING = "WEEK_DEGUMMING"
STAIN = "STAIN"